#!/private/tmp/venv/bin/python
# -*- coding: utf-8 -*-

# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2022.                            (c) 2022.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************

import re
from urllib.parse import unquote
import logging
from xml.etree import ElementTree
import sys
import errno
import time
import base64

logger = logging.getLogger(__name__)

__all__ = ['get_header_filename', 'extract_md5', 'add_md5_header', 'Transfer']


VO_VIEW_DEFAULT = 'ivo://ivoa.net/vospace/core#defaultview'
# CADC specific views
VO_CADC_VIEW_URI = 'ivo://cadc.nrc.ca/vospace/view'
CADC_VO_VIEWS = {'data': '{}#data'.format(VO_CADC_VIEW_URI),
                 'manifest': '{}#manifest'.format(VO_CADC_VIEW_URI),
                 'rss': '{}#rss'.format(VO_CADC_VIEW_URI),
                 'cutout': '{}#cutout'.format(VO_CADC_VIEW_URI)}
SSO_SECURITY_METHODS = {
    'tls-with-certificate': 'ivo://ivoa.net/sso#tls-with-certificate',
    'cookie': 'ivo://ivoa.net/sso#cookie',
    'token': 'vos://cadc.nrc.ca~vospace/CADC/std/Auth#token-1.0'
}


def get_header_filename(headers):
    """
    Extracts the file name from Content-Disposition in the GET response

    Partial implementation of https://tools.ietf.org/html/rfc6266)
    :param headers: HTTP headers of the response
    :return: Name of the file or None if not found
    """
    cd = headers.get('content-disposition', None)
    if not cd:
        return None
    fname = re.findall(r'filename\*=([^;]+)', cd, flags=re.IGNORECASE)
    if not fname:
        fname = re.findall("filename=([^;]+)", cd, flags=re.IGNORECASE)
    if "utf-8''" in fname[0].lower():
        fname = re.sub("utf-8''", '', fname[0], flags=re.IGNORECASE)
        if not isinstance(fname, str):
            # TODO remove - this is just for Python2 support
            fname = unquote(fname.encode('utf-8')).decode('utf-8')
        else:
            fname = unquote(fname)
    else:
        fname = fname[0]
    # clean space and double quotes
    return fname.strip().strip('"')


def extract_md5(headers):
    """ Extracts the md5checksum from the digest header
        :param headers: existing request headers

        :returns string with hex md5 representation if present or None otherwise
    """
    if ('digest' in headers) and (headers['digest'].startswith('md5')):
        value = base64.b64decode(headers['digest'][4:])
        if len(value) == 16:
            return value.hex()
        else:
            #  TODO this is wrong and will be removed eventually after AD is gone
            return value.decode('ascii')
    return None


def add_md5_header(headers, md5_checksum):
    """ Adds the digest header with md5 information to the request headers
        :param headers: existing request headers
        :param md5_checksum: string with hex representation
    """
    headers['digest'] = 'md5={}'.format(
        base64.b64encode(bytes.fromhex(md5_checksum)).decode('ascii'))


class Transfer(object):
    IVOAURL = 'ivo://ivoa.net/vospace/core'
    VOSNS = 'http://www.ivoa.net/xml/VOSpace/v2.0'
    VOSVERSION = '2.1'
    XSINS = 'http://www.w3.org/2001/XMLSchema-instance'
    TYPE = '{{{}}}type'.format(XSINS)
    NODES = '{{{}}}nodes'.format(VOSNS)
    NODE = '{{{}}}node'.format(VOSNS)
    PROTOCOL = '{{{}}}protocol'.format(VOSNS)
    PROPERTIES = '{{{}}}properties'.format(VOSNS)
    PROPERTY = '{{{}}}property'.format(VOSNS)
    ACCEPTS = '{{{}}}accepts'.format(VOSNS)
    PROVIDES = '{{{}}}provides'.format(VOSNS)
    ENDPOINT = '{{{}}}endpoint'.format(VOSNS)
    TARGET = '{{{}}}target'.format(VOSNS)
    DATA_NODE = 'vos:DataNode'
    LINK_NODE = 'vos:LinkNode'
    CONTAINER_NODE = 'vos:ContainerNode'

    DIRECTION_PULL_FROM = "pullFromVoSpace"
    DIRECTION_PUSH_TO = "pushToVoSpace"

    def __init__(self, session):
        """
        Handle transfer related business.  This is here to be reused as needed.
        :param session : requests session to use to communicate with the server
        """
        self.session = session

    def transfer(self, endpoint_url, uri, direction, view=None, cutout=None,
                 security_methods=None, with_uws_job=True):
        """Build the transfer XML document
        :param endpoint_url: URL of the endpoint to POST to
        :param uri: the uri to transfer from or to VOSpace.
        :param direction: is this a pushToVoSpace or a pullFromVoSpace ?
        :param view: which view of the node (data/default/cutout/etc.) is
        being transferred
        :param cutout: a special parameter added to the 'cutout' view
        request. e.g. '[0][1:10,1:10]'
        :param security_methods: the IVOA SSO security methods (see
        vos.SSO_SECURITY_METHODS) that the client
        :param with_uws_job: False if simplified transfer negotiation with no
        UWS jobs (such as in storage inventory)
        intends to use. The service is supposed to return appropriate URLs
        for the security_method. When the security method is not present,
        the service returns pre-authorized URLs that only work for the
        intended purpose, e.g. a pre-authorized PUT URL cannot be used for
        a read (GET or HEAD)

        :raises When a network problem occurs, it raises one of the
        HttpException exceptions declared in the
        cadcutils.exceptions module
        """
        protocol = {
            Transfer.DIRECTION_PULL_FROM: "httpsget",
            Transfer.DIRECTION_PUSH_TO: "httpsput"}

        transfer_xml = ElementTree.Element("vos:transfer")
        transfer_xml.attrib['xmlns:vos'] = Transfer.VOSNS
        transfer_xml.attrib['version'] = Transfer.VOSVERSION
        ElementTree.SubElement(transfer_xml, "vos:target").text = uri
        ElementTree.SubElement(transfer_xml, "vos:direction").text = \
            direction

        if view == 'move':
            ElementTree.SubElement(transfer_xml,
                                   "vos:keepBytes").text = "false"
        else:
            if view == 'defaultview':
                ElementTree.SubElement(transfer_xml, "vos:view").attrib[
                    'uri'] = VO_VIEW_DEFAULT
            elif view is not None:
                vos_view = ElementTree.SubElement(transfer_xml, "vos:view")
                vos_view.attrib['uri'] = CADC_VO_VIEWS[view]
                if cutout is not None and view == 'cutout':
                    param = ElementTree.SubElement(vos_view, "vos:param")
                    param.attrib['uri'] = CADC_VO_VIEWS[view]
                    param.text = cutout
            protocol_element = ElementTree.SubElement(transfer_xml,
                                                      "vos:protocol")
            protocol_element.attrib['uri'] = "{0}#{1}".format(
                Transfer.IVOAURL, protocol[direction])

            if security_methods:
                for sm in security_methods:
                    if sm not in SSO_SECURITY_METHODS.values():
                        raise AttributeError(
                            'Invalid security method {}. Supported values: {}'.
                            format(sm, SSO_SECURITY_METHODS.values))
                    security_element = ElementTree.SubElement(
                        protocol_element, "vos:securityMethod")
                    security_element.attrib['uri'] = sm
            if security_methods:
                # add the default anon method
                protocol_element = ElementTree.SubElement(transfer_xml,
                                                          "vos:protocol")
                protocol_element.attrib['uri'] = "{0}#{1}".format(
                    Transfer.IVOAURL, protocol[direction])

        logging.debug(ElementTree.tostring(transfer_xml))
        logging.debug("Sending to : {}".format(endpoint_url))

        data = ElementTree.tostring(transfer_xml)
        resp = self.session.post(
            endpoint_url, data=data, allow_redirects=False,
            headers={'Content-Type': 'text/xml'})

        logging.debug("{0}".format(resp))
        logging.debug("{0}".format(resp.content))
        while resp.status_code == 303:
            goto_url = resp.headers.get('Location', None)

            if self.session.auth is not None and \
                    "auth" not in goto_url:
                goto_url = goto_url.replace('/vospace/', '/vospace/auth/')

            logging.debug(
                'Got back from transfer URL: {}'.format(goto_url))
            # for get or put we need the protocol value
            resp = self.session.get(goto_url, allow_redirects=False)
        if resp.status_code == 200:
            transfer_url = str(resp.url)
            if view == 'move':
                return transfer_url
            if with_uws_job:
                # check the status of the job first
                self.check_job_error(
                    str.replace(transfer_url, 'xfer', 'transfers'),
                    str(uri), True)
            xml_string = resp.content
            logging.debug('Transfer Document:{}'.format(xml_string))
            transfer_document = ElementTree.fromstring(xml_string)
            logging.debug(
                "XML version: {0}".format(
                    ElementTree.tostring(transfer_document)))
            all_protocols = transfer_document.findall(Transfer.PROTOCOL)
            if all_protocols is None or not len(all_protocols) > 0:
                raise RuntimeError(
                    "BUG: No protocol/endpoint returned for transfer URL {}".
                    format(transfer_url))
        elif resp.status_code == 404:
            raise OSError(resp.status_code,
                          "File not found: {0}".format(uri))
        else:
            raise OSError(resp.status_code,
                          "Failed to get transfer service response.")

        result = []
        for protocol in all_protocols:
            for node in protocol.findall(Transfer.ENDPOINT):
                result.append(node.text)
        return result

    def _get_phase(self, phase_url):
        response = self.session.get(phase_url, allow_redirects=True)
        response.raise_for_status()
        return response.text

    def get_transfer_error(self, url, uri):
        """Follow a transfer URL to completion and returns the error if it
        fails
        :param url: The URL of the transfer request that had the error.
        :param uri: The uri that we were trying to transfer (get or put).

        :raises When a network problem occurs, it raises one of the
        HttpException exceptions declared in the
        cadcutils.exceptions module. Returns None if job completes successfully
        """
        job_url = str.replace(url, "/results/transferDetails", "")

        try:
            phase_url = job_url + "/phase"
            sleep_time = 1
            roller = ('\\', '-', '/', '|', '\\', '-', '/', '|')
            phase = self._get_phase(phase_url)
            # do not remove the line below. It is used for testing
            logging.debug("Job URL: " + job_url + "/phase")
            while phase in ['PENDING', 'QUEUED', 'EXECUTING', 'UNKNOWN']:
                # poll the job. Sleeping time in between polls is doubling
                # each time until it gets to 32sec
                total_time_slept = 0
                if sleep_time <= 32:
                    sleep_time *= 2
                slept = 0
                if logger.getEffectiveLevel() == logging.INFO:
                    while slept < sleep_time:
                        sys.stdout.write(
                            '\r{} {}'.format(phase, roller[total_time_slept %
                                                           len(roller)]))
                        sys.stdout.flush()
                        slept += 1
                        total_time_slept += 1
                        time.sleep(1)
                    sys.stdout.write("\r                    \n")
                else:
                    time.sleep(sleep_time)
                phase = self._get_phase(phase_url)
                logging.debug(
                    'Async transfer Phase for url {}: {}'.format(url, phase))
        except KeyboardInterrupt:
            # abort the job when receiving a Ctrl-C/Interrupt from the client
            logging.error("Received keyboard interrupt")
            self.session.post(
                job_url + "/phase", allow_redirects=False, data="PHASE=ABORT",
                headers={'Content-type': 'application/x-www-form-urlencoded'})
            raise KeyboardInterrupt
        phase = self._get_phase(phase_url)
        logger.debug("Phase:  {0}".format(phase))
        if phase in ['COMPLETED']:
            return False
        if phase in ['HELD', 'SUSPENDED', 'ABORTED']:
            # re-queue the job and continue to monitor for completion.
            raise OSError("UWS status: {0}".format(phase), errno.EFAULT)
        return self.check_job_error(url, uri)

    def check_job_error(self, url, uri, check_phase=False):
        """
        Checks whether a job is an error state and raises the appropriate
        exception. Job does not have to be completed.
        :param url: job url
        :param uri: vospace artifac
        :param check_phase: True if job phase needs to be read from server,
        False otherwise
        :return: Raises an exception if job failed.
        """
        error_codes = {'NodeNotFound': errno.ENOENT,
                       'RequestEntityTooLarge': errno.E2BIG,
                       'PermissionDenied': errno.EACCES,
                       'OperationNotSupported': errno.EOPNOTSUPP,
                       'InternalFault': errno.EFAULT,
                       'ProtocolNotSupported': errno.EPFNOSUPPORT,
                       'ViewNotSupported': errno.ENOSYS,
                       'InvalidArgument': errno.EINVAL,
                       'InvalidURI': errno.EFAULT,
                       'TransferFailed': errno.EIO,
                       'DuplicateNode.': errno.EEXIST,
                       'NodeLocked': errno.EPERM}
        job_url = str.replace(str(url), "/results/transferDetails", "")
        if check_phase:
            phase = self._get_phase(job_url + "/phase")
            if phase not in ['ERROR']:
                return False
        error_url = job_url + "/error"
        error_message = self.session.get(error_url).text
        logger.debug(
            "Got transfer error {0} on URI {1}".format(error_message, uri))
        # Check if the error was that the link type is unsupported and try and
        # follow that link.
        target = re.search('Unsupported link target:(?P<target> .*)$',
                           error_message)
        if target is not None:
            return target.group('target').strip()
        raise OSError(error_codes.get(error_message, errno.EFAULT),
                      '{0}: {1}'.format(uri, error_message))



"""
The cadcutils.net package contains network-related utilities commonly use in
CADC applications.

The main functions/classes are as follows (Note that they can be imported
directly from cadcutils.net):

    - get_cert: retrieves a X509 proxy certificate from the Credential
    Delegation Protocol (CDP)
                Web Service
    - Subject: container class for user credentials.
    - BaseWsClient: this is the base class for interacting with a Web service
    - GroupsClient: class used to interact with the CADC Group Management
    Service (GMS)

Note that the library also defines the entry_point for the cadc-get-cert at
cadcutils.net.auth:get_cert_main and one for cadc-groups at
cadcutils.net.groups_client.main_app

Examples of usage:
1. Instantiate Groups and use it to access GMS.

One mandatory argument that the CadcTapClient constructor takes is a
cadcutils.net.Subject that holds the user credentials.

Example:
   from cadcutils import net

   client = net.GroupsClient(net.Subject(certificate=<certificate-location>))
   print(client.get_group(<group_id>)
   ...
   # other client calls

2. Invoke the cadc-groups entry point function. This is the function that
is used to generate the cadc-groups application

Example:
   from cadcutils.net.groups_client import main_app
   import sys

   sys.argv = ['cadc-groups', '-h']
   main_app()

3. Invoke the cadc-groups as an external command
Example:
   import os
   os.system('cadc-groups --help')

Method 1. is the recommended method as it does not require forking external
process and also allows trapping the exceptions and reacting according to the
type of the error. Method 2 also works but the sys.exit needs to be trapped in
order to prevent the script from quiting. Method 3, while simple, must rely on
inter processes communication to determine the result of running the command.

"""

from .auth import *  # noqa
from .ws import *  # noqa
from .netutils import *  # noqa
from .group import *  # noqa
from .groups_client import *  # noqa



# -*- coding: utf-8 -*-

# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2022.                            (c) 2022.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************

"""
Module that contains functionality related to authentication. The
main tools in this module are get_cert that interacts with the
CADC Credential Delegation Protocol Web Service to return a proxy
X509 certificate and Subject that incapsulates the credentials of
a user.
"""

import getpass
import netrc as netrclib
import os
import signal
import sys
import html2text

from cadcutils.net import ws
from cadcutils import util, exceptions, version

CRED_RESOURCE_ID = 'ivo://cadc.nrc.ca/cred'
CRED_PROXY_FEATURE_ID = 'ivo://ivoa.net/std/CDP#proxy-1.0'

__all__ = ['get_cert', 'Subject']

# these are the security methods currently supported
SECURITY_METHODS_IDS = {
    'certificate': 'ivo://ivoa.net/sso#tls-with-certificate',
    'basic': 'ivo://ivoa.net/sso#BasicAA',
    'cookie': 'ivo://ivoa.net/sso#cookie'}

SUPPORTED_SERVER_VERSIONS = {'cred': '2.0'}

logger = util.get_logger(__name__)


class CookieInfo(object):
    """
    Class to store information regarding an HTTP Cookie
    """
    def __init__(self, domain, name, value):
        self.domain = domain
        self.name = name
        self.value = value


class Subject(object):
    """
    Class that stores user authentication information to be used for accessing
        distributed resources over HTTP. For now, the supported credentials
        are:
        X509 certificate (via a proxy certificate file) or basic HTTP
        user/password (via netrc file that stores user/password or simply user
        name. Note that in the later case the library might prompt for the
        password before connecting.
    """

    def __init__(self, username=None, certificate=None, netrc=False):
        """
            The subject is anonymous if neither of this arguments is set
        :param username: user name
        :param certificate: name of the X509 certificate file
        :param netrc: use information from .netrc. Value can be True (use
        default $HOME/.netrc) or the name of the netrc file to use.
        """
        self.username = username
        self._hosts_auth = {}
        self._certificate = None
        self.certificate = certificate
        self._netrc = False
        self.netrc = netrc
        self._cookies = []

    @property
    def certificate(self):
        """
        X509 certificate associated with the subject
        :return: name of the X509 certificate file
        """
        return self._certificate

    @certificate.setter
    def certificate(self, value):
        if value is not None:
            if not os.path.isfile(value):
                raise ValueError('Certificate file {} not found'.format(value))
            self._certificate = value

    @property
    def netrc(self):
        """
        netrc file containing user/passwds credentials (if used by subject)
        :return: Name of netrc file
        """
        return self._netrc

    @netrc.setter
    def netrc(self, value):
        if value is not False:
            hosts = netrclib.netrc(value if value is not True else None).hosts
            if value is True:
                self._netrc = os.path.join(os.environ['HOME'], ".netrc")
            else:
                self._netrc = value
            for host_name in hosts:
                self._hosts_auth[host_name] = (hosts[host_name][0],
                                               hosts[host_name][2])

    @property
    def anon(self):
        """
        Is this an anonymous subject (has no authentication means)?
        :return:
        """
        return (self.certificate is None) and (self.netrc is False) and\
               (self.username is None)

    @property
    def cookies(self):
        return self._cookies

    @staticmethod
    def from_cmd_line_args(args):
        """
        Instantiates a subject based on attributes of a command line. It
        works with the base parser in cadcutils and uses the following command
        line arguments:
            args.user: username
            args.cert: x509 certificate location
            args.n: use netrc files for authentication info
            args.netrc_file: use this netrc file for authentication info
        :param args: argparse command line arguments
        :return: corresponding subject
        """
        return Subject(username=args.user, certificate=args.cert,
                       netrc=(args.netrc_file if args.netrc_file
                              is not None else args.n))

    def get_auth(self, realm):
        """
        Returns a user/password touple for the given realm. Note that this
        function prompts for the password on stdout when the username of the
        subject is known but no corresponding password can be found

        :param realm: realm for the authentication
        :return: (username, password) touple or None if subject is anonymous
        or password not found.
        """
        if self.anon:
            return None
        if self.username is None:
            if realm in self._hosts_auth:
                return self._hosts_auth[realm]
            else:
                msg = 'No user/password for {}'.format(realm)
                if self.netrc is not False:
                    msg = '{} in {}'.format(msg,
                                            self.netrc if self.netrc
                                            is not True else '$HOME/.netrc')
                logger.debug(msg)
                return None
        else:
            if realm in self._hosts_auth \
                    and self.username == self._hosts_auth[realm][0]:
                return self._hosts_auth[realm]
            sys.stdout.write("{}@{}\n".format(self.username, realm))
            sys.stdout.flush()
            pswd = getpass.getpass().strip().strip('\n')
            if not pswd:
                raise ValueError('Password cannot be empty')
            self._hosts_auth[realm] = (self.username, pswd)
            sys.stdout.write("\n")
            sys.stdout.flush()
            return self._hosts_auth[realm]

    def get_security_methods(self):
        """
        returns the security method IDs that this subject is authentication
        for. The order of the returned methods is one that it is preferred:
        certificate, basic and anon.

        Note: a realm would be required here as some methods (cookie, basic)
        are bound to a realm
        :return: list of security method IDs
        """
        sms = []
        if self.certificate is not None:
            sms.append(SECURITY_METHODS_IDS['certificate'])
        if self.cookies:
            sms.append(SECURITY_METHODS_IDS['cookie'])
        if (self.netrc is not False) or (self.username is not None):
            sms.append(SECURITY_METHODS_IDS['basic'])
        return sms


def get_cert(subject, days_valid=None, host=None, insecure=False):
    """Access the CADC Certificate Delegation Protocol (CDP) server and
       retrieve a X509 proxy certificate.

    :param: subject: subject performing the action
    :ptype: cadcutils.subject
    :param: host: name of the host (overrides the host returned by the service
    registry)
    :param: days_valid: number of days the proxy certificate is valid for
    :ptype daysValid: int
    :param insecure: Allow insecure server connections over SSL (testing)
    :ptype insecure: boolean

    :return content of the certificate

    """
    params = {}
    if days_valid is not None:
        params['daysValid'] = int(days_valid)
    util.check_version(version=version.version)
    client = ws.BaseWsClient(CRED_RESOURCE_ID, subject,
                             agent="cadc-get-cert/" + version.version, retry=True,
                             host=host, server_versions=SUPPORTED_SERVER_VERSIONS,
                             insecure=insecure)
    response = client.get((CRED_PROXY_FEATURE_ID, None), params=params)
    return response.text


def get_cert_main():
    """ Client to download an X509 certificate and save it in users home
    directory"""

    def _signal_handler(signal, frame):
        sys.stderr.write("\n")
        sys.exit(-1)

    signal.signal(signal.SIGINT, _signal_handler)

    parser = util.get_base_parser(subparsers=False, version=version.version,
                                  default_resource_id=CRED_RESOURCE_ID,
                                  auth_required=True)
    parser.description = ('Retrieve a security certificate for interaction '
                          'with a Web service such as VOSpace. Certificate '
                          'will be valid for days-valid and stored as local '
                          'file cert_filename.')
    parser.add_argument('--cert-filename',
                        default=os.path.join(os.getenv('HOME', '/tmp'),
                                             '.ssl/cadcproxy.pem'),
                        help=('filesystem location to store the proxy '
                              'certificate. (default: {})'.
                              format(os.path.join(os.getenv('HOME', '/tmp'),
                                                  '.ssl/cadcproxy.pem'))))
    parser.add_argument('--days-valid', type=int, default=10,
                        help='number of days the certificate should be valid.')

    args = parser.parse_args()

    dirname = os.path.dirname(args.cert_filename)
    if dirname:
        try:
            os.makedirs(dirname)
        except OSError as oex:
            if os.path.isdir(dirname):
                pass
            elif oex.errno == 20 or oex.errno == 17:
                sys.stderr.write("%s : %s\n" % (str(oex), dirname))
                sys.stderr.write("Expected %s to be a directory.\n" % dirname)
                sys.exit(oex.errno)
            else:
                raise oex

    try:
        subject = Subject.from_cmd_line_args(args)
        cert = get_cert(subject, days_valid=args.days_valid, host=args.host,
                        insecure=args.insecure)
        with open(args.cert_filename, 'w') as w:
            w.write(cert)
        if not args.quiet:
            print('DONE. {} day certificate saved in {}'.format(
                args.days_valid, args.cert_filename))
    except exceptions.UnauthorizedException:
        # unauthorized
        sys.stderr.write('FAILED: invalid username/password combination')
    except Exception as ex:
        sys.stderr.write("FAILED to retrieve {} day certificate\n".format(
            args.days_valid))
        sys.stderr.write('{}'.format(html2text.html2text(str(ex))))
        return getattr(ex, 'errno', 1)



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2017.                            (c) 2017.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************

import unittest
import pytest

from cadcutils.net import wscapabilities


class TestWsCapabilities(unittest.TestCase):
    def test_parsing(self):
        # tests some error cases
        cr = wscapabilities.CapabilitiesReader()
        with self.assertRaises(ValueError):
            cr.parsexml('blah')

        with self.assertRaisesRegex(
                ValueError,
                'BUG reading remote service info '
                r'\(capabilities\) - No actual capabilities found'):
            cr.parsexml('<capabilities></capabilities>')

        with self.assertRaisesRegex(ValueError,
                                    r'Cannot read service info \(capabilities\). '
                                    'Capability standard ID is invalid URL: None'):
            cr.parsexml(
                '<capabilities><capability></capability></capabilities>')

        with self.assertRaisesRegex(ValueError,
                                    r'Cannot read service info \(capabilities\). '
                                    'Capability standard ID is invalid URL: abc'):
            cr.parsexml(
                '<capabilities><capability '
                'standardID="abc"></capability></capabilities>')

        with self.assertRaisesRegex(ValueError,
                                    r'BUG reading service info \(capabilities\). '
                                    'No interfaces found for capability '
                                    'ivo://provider/service'):
            cr.parsexml(
                '<capabilities><capability '
                'standardID="ivo://provider/service">'
                '</capability></capabilities>')

        with self.assertRaisesRegex(
                ValueError,
                r'Cannot read service info \(capabilities\). '
                'No accessURL for ivo://provider/service'):
            cr.parsexml(
                '<capabilities '
                'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
                '<capability standardID="ivo://provider/service">'
                '<interface xsi:type="vs:ParamHTTP">'
                '</interface></capability></capabilities>')

        with self.assertRaisesRegex(
                ValueError,
                r'Cannot read service info \(capabilities\). '
                'No accessURL for ivo://provider/service'):
            cr.parsexml(
                '<capabilities xmlns:xsi="ns"><capability '
                'standardID="ivo://provider/service">'
                '<interface></interface><accessURL>'
                '</accessURL></capability></capabilities>')

        with self.assertRaisesRegex(ValueError,
                                    r'Cannot read service info \(capabilities\). '
                                    'No accessURL for ivo://provider/service'):
            cr.parsexml(
                '<capabilities '
                'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
                '<capability '
                'standardID="ivo://provider/service">'
                '<interface xsi:type="vs:ParamHTTP">'
                '<accessURL standardID="abc">'
                '</accessURL></interface></capability></capabilities>')

        # simplest capabilities document that parses successfully
        caps = cr.parsexml(
            '<capabilities '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            '<capability standardID="ivo://provider/service">'
            '<interface xsi:type="vs:ParamHTTP">'
            '<accessURL>http://someurl/somepath'
            '</accessURL></interface></capability></capabilities>')

        assert 'http://someurl/somepath' == \
               caps.get_access_url('ivo://provider/service', None)

        # multiple interfaces => go to the first one
        caps = cr.parsexml(
            '<capabilities '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            '<capability standardID="ivo://provider/service">'
            '<interface xsi:type="vs:ParamHTTP">'
            '<accessURL>https://someurl/somepath'
            '</accessURL></interface>'
            '<interface xsi:type="vs:ParamHTTP">'
            '<accessURL>http://someurl/somepath'
            '</accessURL></interface></capability></capabilities>')

        assert 'https://someurl/somepath' == \
               caps.get_access_url('ivo://provider/service', None)

        # multiple interfaces => https url preferred even when change order
        caps = cr.parsexml(
            '<capabilities '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            '<capability standardID="ivo://provider/service">'
            '<interface  xsi:type="vs:ParamHTTP">'
            '<accessURL>https://someurl/somepath'
            '</accessURL></interface>'
            '<interface xsi:type="vs:ParamHTTP">'
            '<accessURL>http://someurl/somepath'
            '</accessURL></interface></capability></capabilities>')

        assert 'https://someurl/somepath' == \
               caps.get_access_url('ivo://provider/service', None)

        # multipel interfaces => first one chosen when both http or first one
        # is already https
        caps = cr.parsexml(
            '<capabilities '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            '<capability standardID="ivo://provider/service">'
            '<interface  xsi:type="vs:ParamHTTP">'
            '<accessURL>https://someurl1/somepath'
            '</accessURL></interface>'
            '<interface xsi:type="vs:ParamHTTP">'
            '<accessURL>https://someurl2/somepath'
            '</accessURL></interface></capability></capabilities>')

        assert 'https://someurl1/somepath' == \
               caps.get_access_url('ivo://provider/service', None)

        caps = cr.parsexml(
            '<capabilities '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            '<capability standardID="ivo://provider/service">'
            '<interface xsi:type="vs:ParamHTTP">'
            '<accessURL>http://someurl1/somepath'
            '</accessURL></interface><interface xsi:type="vs:ParamHTTP">'
            '<accessURL>http://someurl2/somepath'
            '</accessURL></interface></capability></capabilities>')

        assert 'http://someurl1/somepath' == \
               caps.get_access_url('ivo://provider/service', None)

        # add security method
        caps = cr.parsexml(
                '<capabilities '
                'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
                '<capability '
                'standardID="ivo://provider/service">'
                '<interface xsi:type="vs:ParamHTTP">'
                '<accessURL>http://someurl/somepath'
                '</accessURL><securityMethod></securityMethod>'
                '</interface></capability></capabilities>')

        assert 'http://someurl/somepath' == \
            caps.get_access_url('ivo://provider/service', None)

        # multiple security methods TAP1.0 style
        caps = cr.parsexml(
            '<capabilities '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            '<capability standardID="ivo://provider/service">'
            '<interface xsi:type="vs:ParamHTTP">'
            '<accessURL>http://someurl/somepath1'
            '</accessURL><securityMethod standardID="ivo://cadc/mymethod1">'
            '</securityMethod></interface>'
            '<interface xsi:type="vs:ParamHTTP">'
            '<accessURL>http://someurl/somepath2'
            '</accessURL><securityMethod standardID="ivo://cadc/mymethod2">'
            '</securityMethod></interface>'
            '</capability></capabilities>')
        assert 'http://someurl/somepath1' == \
               caps.get_access_url('ivo://provider/service',
                                   ['ivo://cadc/mymethod1'])
        assert 'http://someurl/somepath2' == \
               caps.get_access_url('ivo://provider/service',
                                   ['ivo://cadc/mymethod2'])
        with pytest.raises(ValueError):
            assert caps.get_access_url('ivo://provider/service', None)

        # multiple security methods TAP1.1 style
        caps = cr.parsexml(
            '<capabilities '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            '<capability standardID="ivo://provider/service">'
            '<interface xsi:type="vs:ParamHTTP">'
            '<accessURL>http://someurl/somepath'
            '</accessURL>'
            '<securityMethod standardID="ivo://cadc/mymethod1">'
            '</securityMethod>'
            '<securityMethod standardID="ivo://cadc/mymethod2">'
            '</securityMethod></interface>'
            '</capability></capabilities>')
        assert 'http://someurl/somepath' == \
               caps.get_access_url('ivo://provider/service',
                                   ['ivo://cadc/mymethod1'])
        assert 'http://someurl/somepath' == \
               caps.get_access_url('ivo://provider/service',
                                   ['ivo://cadc/mymethod2'])
        with pytest.raises(ValueError):
            assert caps.get_access_url('ivo://provider/service', None)

        # similar with the above test except anonymous security method
        # accepted
        caps = cr.parsexml(
            '<capabilities '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            '<capability standardID="ivo://provider/service">'
            '<interface xsi:type="vs:ParamHTTP">'
            '<accessURL>http://someurl/somepath'
            '</accessURL>'
            '<securityMethod standardID="ivo://cadc/mymethod1">'
            '</securityMethod>'
            '<securityMethod/>'
            '<securityMethod standardID="ivo://cadc/mymethod2">'
            '</securityMethod></interface>'
            '</capability></capabilities>')
        assert 'http://someurl/somepath' == \
               caps.get_access_url('ivo://provider/service',
                                   ['ivo://cadc/mymethod1'])
        assert 'http://someurl/somepath' == \
               caps.get_access_url('ivo://provider/service',
                                   ['ivo://cadc/mymethod2'])
        assert 'http://someurl/somepath' == \
               caps.get_access_url('ivo://provider/service', None)
        # unlisted security method would be treated as anonymous
        assert 'http://someurl/somepath' == \
               caps.get_access_url('ivo://provider/service',
                                   ['ivo://cadc/mymethod3'])

        # test only anonymous TAP1.0 style
        caps = cr.parsexml(
            '<capabilities '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            '<capability standardID="ivo://provider/service">'
            '<interface xsi:type="vs:ParamHTTP">'
            '<accessURL>http://someurl/somepath'
            '</accessURL>'
            '</interface>'
            '</capability></capabilities>')
        assert 'http://someurl/somepath' == \
               caps.get_access_url('ivo://provider/service', None)
        with pytest.raises(ValueError):
            assert caps.get_access_url('ivo://provider/service', "anymethod")

        # test only anonymous TAP1.1 style
        caps = cr.parsexml(
            '<capabilities '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            '<capability standardID="ivo://provider/service">'
            '<interface xsi:type="vs:ParamHTTP">'
            '<accessURL>http://someurl/somepath'
            '</accessURL>'
            '<securityMethod />'
            '</interface>'
            '</capability></capabilities>')
        assert 'http://someurl/somepath' == \
               caps.get_access_url('ivo://provider/service', None)
        with pytest.raises(ValueError):
            assert caps.get_access_url('ivo://provider/service', "anymethod")
